#!/bin/python3

"""
About this Class
"""

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
warnings.simplefilter(action='ignore', category=UserWarning)

from subject.subject import Subject
from group_level.group import GroupLevel
from aggregator.aggregator import Aggregator